# -*- coding: utf-8 -*-
# Copyright 2015-2016 Yelp
# Copyright 2019 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parse the log4j syslog format used by Hadoop."""
import re
from collections import namedtuple

# log line format output by hadoop jar command
_HADOOP_LOG4J_LINE_RE = re.compile(
    r"^\s*(?P<timestamp>\d\d\/\d\d\/\d\d \d\d\:\d\d\:\d\d)"
    r"\s+(?P<level>[A-Z]+)"
    r"\s+(?P<logger>\S+)"
    r"(\s+\((?P<thread>.*?)\))?"
    r"( - ?|: ?)"
    r"(?P<message>.*?)$"
)

# log line format output to Hadoop syslog
_HADOOP_LOG4J_LINE_ALTERNATE_RE = re.compile(
    r"^\s*(?P<timestamp>\d\d\/\d\d\/\d\d \d\d\:\d\d\:\d\d)"
    r"\s+(?P<level>[A-Z]+)"
    r"(\s+\[(?P<thread>.*?)\])"
    r"\s+(?P<logger>\S+)"
    r"(\s+\((?P<caller_location>\S+)\))?"
    r"( - ?|: ?)"
    r"(?P<message>.*?)$"
)


class Log4jRecord(
    namedtuple(
        "_Log4jRecord", "caller_location level logger message num_lines start_line thread timestamp"
    )
):
    r"""Represents a Log4J log record.

    caller_location -- e.g. 'YarnClientImpl.java:submitApplication(251)'
    level -- e.g. 'INFO'
    logger -- e.g. 'amazon.emr.metrics.MetricsSaver'
    message -- the actual message. If this is a multi-line message (e.g.
        for counters), the lines will be joined by '\n'
    num_lines -- how many lines made up the message
    start_line -- which line the message started on (0-indexed)
    thread -- e.g. 'main'. Defaults to ''
    timestamp -- unparsed timestamp, e.g. '15/12/07 20:49:28'
    """

    def __new__(
        cls, caller_location, level, logger, message, num_lines, start_line, thread, timestamp
    ):
        return super(Log4jRecord, cls).__new__(
            cls, caller_location, level, logger, message, num_lines, start_line, thread, timestamp
        )

    @staticmethod
    def fake_record(line, line_num):
        """Used to represent a leading Log4J line that doesn't conform to the regular expressions we
        expect.
        """
        return Log4jRecord(
            caller_location="",
            level="",
            logger="",
            message=line,
            num_lines=1,
            start_line=line_num,
            thread="",
            timestamp="",
        )


def parse_hadoop_log4j_records(lines):
    r"""Parse lines from a hadoop log into log4j records.

    Yield Log4jRecords.

    Lines will be converted to unicode, and trailing \r and \n will be stripped
    from lines.

    Also yields fake records for leading non-log4j lines (trailing non-log4j
    lines are assumed to be part of a multiline message if not pre-filtered).
    """
    last_record = None
    line_num = 0

    for line_num, line in enumerate(lines.split("\n")):
        # convert from bytes to unicode, if needed, and strip trailing newlines
        line = line.rstrip("\r\n")

        m = _HADOOP_LOG4J_LINE_RE.match(line) or _HADOOP_LOG4J_LINE_ALTERNATE_RE.match(line)

        if m:
            if last_record:
                last_record = last_record._replace(num_lines=line_num - last_record.start_line)
                yield last_record

            matches = m.groupdict()

            last_record = Log4jRecord(
                caller_location=matches.get("caller_location", ""),
                level=matches["level"],
                logger=matches["logger"],
                message=matches["message"],
                num_lines=1,
                start_line=line_num,
                thread=matches.get("thread", ""),
                timestamp=matches["timestamp"],
            )
        else:
            # add on to previous record
            if last_record:
                last_record = last_record._replace(message=last_record.message + "\n" + line)
            else:
                yield Log4jRecord.fake_record(line, line_num)

    if last_record:
        last_record = last_record._replace(num_lines=line_num + 1 - last_record.start_line)
        yield last_record



# This file duplicated from the Yelp MRJob project:
#
#   https://github.com/Yelp/mrjob
#
#
# Copyright 2009-2013 Yelp, David Marin
# Copyright 2015 Yelp
# Copyright 2017 Yelp
# Copyright 2018 Contributors
# Copyright 2019 Yelp and Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrappers for gracefully retrying on error."""
import logging
import time
from functools import partial

log = logging.getLogger(__name__)


_DEFAULT_BACKOFF = 15
_DEFAULT_MULTIPLIER = 1.5
_DEFAULT_MAX_TRIES = 10
_DEFAULT_MAX_BACKOFF = 1200  # 20 minutes


class RetryWrapper:
    """Handle transient errors, with configurable backoff.

    This class can wrap any object. The wrapped object will behave like
    the original one, except that if you call a function and it raises a
    retriable exception, we'll back off for a certain number of seconds
    and call the function again, until it succeeds or we get a non-retriable
    exception.
    """

    def __init__(
        self,
        wrapped,
        retry_if,
        backoff=_DEFAULT_BACKOFF,
        multiplier=_DEFAULT_MULTIPLIER,
        max_tries=_DEFAULT_MAX_TRIES,
        max_backoff=_DEFAULT_MAX_BACKOFF,
        unwrap_methods=(),
    ):
        """
        Wrap the given object.

        :param wrapped: the object to wrap
        :param retry_if: a method that takes an exception, and returns whether
                         we should retry
        :type backoff: float
        :param backoff: the number of seconds to wait the first time we get a
                        retriable error
        :type multiplier: float
        :param multiplier: if we retry multiple times, the amount to multiply
                           the backoff time by every time we get an error
        :type max_tries: int
        :param max_tries: how many tries we get. ``0`` means to keep trying
                          forever
        :type max_backoff: float
        :param max_backoff: cap the backoff at this number of seconds
        :type unwrap_methods: sequence
        :param unwrap_methods: names of methods to call with this object as
                               *self* rather than retrying on transient
                               errors (e.g. methods that return a paginator)
        """
        self.__wrapped = wrapped

        self.__retry_if = retry_if

        self.__backoff = backoff
        if self.__backoff <= 0:
            raise ValueError("backoff must be positive")

        self.__multiplier = multiplier
        if self.__multiplier < 1:
            raise ValueError("multiplier must be at least one!")

        self.__max_tries = max_tries

        self.__max_backoff = max_backoff

        self.__unwrap_methods = set(unwrap_methods)

    def __getattr__(self, name):
        """The glue that makes functions retriable, and returns other
        attributes from the wrapped object as-is.
        """
        x = getattr(self.__wrapped, name)

        if name in self.__unwrap_methods:
            return partial(x.__func__, self)
        elif hasattr(x, "__call__"):
            return self.__wrap_method_with_call_and_maybe_retry(x)
        else:
            return x

    def __wrap_method_with_call_and_maybe_retry(self, f):
        """Wrap method f in a retry loop."""

        def call_and_maybe_retry(*args, **kwargs):
            backoff = self.__backoff
            tries = 0

            while not self.__max_tries or tries < self.__max_tries:
                try:
                    return f(*args, **kwargs)
                except Exception as ex:
                    if self.__retry_if(ex) and (
                        tries < self.__max_tries - 1 or not self.__max_tries
                    ):
                        log.info("Got retriable error: %r" % ex)
                        log.info("Backing off for %.1f seconds" % backoff)
                        time.sleep(backoff)
                        tries += 1
                        backoff *= self.__multiplier
                        backoff = min(backoff, self.__max_backoff)
                    else:
                        raise

        # pretend to be the original function
        call_and_maybe_retry.__doc__ = f.__doc__

        if hasattr(f, "__name__"):
            call_and_maybe_retry.__name__ = f.__name__

        return call_and_maybe_retry



from .resources import (
    FakeRedshiftResource as FakeRedshiftResource,
    RedshiftError as RedshiftError,
    RedshiftResource as RedshiftResource,
    fake_redshift_resource as fake_redshift_resource,
    redshift_resource as redshift_resource,
)
